#include "ArvizIntrospector.h"

#include <range/v3/view/enumerate.hpp>

#include <Eigen/Geometry>

#include <SimoxUtility/color/Color.h>
#include <VirtualRobot/Robot.h>

#include "RobotAPI/components/ArViz/Client/Elements.h"
#include "RobotAPI/components/ArViz/Client/elements/Path.h"

#include "Navigation/libraries/core/types.h"
#include "Navigation/libraries/server/StackResult.h"

namespace armarx::nav::server
{
    ArvizIntrospector::ArvizIntrospector(armarx::viz::Client arviz,
                                         const VirtualRobot::RobotPtr& robot) :
        arviz(arviz), robot(robot)
    {
    }

    void
    ArvizIntrospector::onStackResult(const StackResult& result)
    {
        // if (result.globalPlannerResult)
        // {
        //     drawGlobalTrajectory(*result.globalPlannerResult);
        // }

        // if (result.localTrajectory)
        // {
        //     drawLocalTrajectory(*result.localTrajectory);
        // }

        // if (result.controlVelocity)
        // {
        //     drawRawVelocity(*result.controlVelocity);
        // }

        // if (result.safeVelocity)
        // {
        //     drawSafeVelocity(*result.safeVelocity);
        // }

        // arviz.commit(layers);

        // layers.clear();
    }

    void
    ArvizIntrospector::drawGlobalTrajectory(const core::Trajectory& trajectory)
    {
        auto layer = arviz.layer("global_planner");

        layer.add(
            viz::Path("path").points(trajectory.positions()).color(simox::color::Color::blue()));

        for (const auto& [idx, tp] : trajectory.points() | ranges::views::enumerate)
        {
            constexpr float scale = 150;

            const Eigen::Vector3f target =
                scale * tp.waypoint.pose.linear() * Eigen::Vector3f::UnitY();

            layer.add(
                viz::Arrow("velocity_" + std::to_string(idx))
                    .fromTo(tp.waypoint.pose.translation(), tp.waypoint.pose.translation() + target)
                    .color(simox::color::Color::blue()));
        }


        layers.emplace_back(std::move(layer));
    }

    void
    ArvizIntrospector::drawLocalTrajectory(const core::Trajectory& trajectory)
    {
        auto layer = arviz.layer("local_planner");

        layer.add(viz::Path("path").points(trajectory.positions()).color(simox::Color::blue()));

        layers.emplace_back(std::move(layer));
    }

    void
    ArvizIntrospector::drawRawVelocity(const core::Twist& twist)
    {
        auto layer = arviz.layer("trajectory_controller");

        layer.add(viz::Arrow("linear_velocity")
                      .fromTo(robot->getGlobalPosition(),
                              core::Pose(robot->getGlobalPose()) * twist.linear)
                      .color(simox::Color::orange()));

        layers.emplace_back(std::move(layer));
    }

    void
    ArvizIntrospector::drawSafeVelocity(const core::Twist& twist)
    {
        auto layer = arviz.layer("safety_controller");

        layer.add(viz::Arrow("linear_velocity")
                      .fromTo(robot->getGlobalPosition(),
                              core::Pose(robot->getGlobalPose()) * twist.linear)
                      .color(simox::Color::green()));

        layers.emplace_back(std::move(layer));
    }

    ArvizIntrospector::ArvizIntrospector(ArvizIntrospector&& other) noexcept :
        arviz{other.arviz}, robot{other.robot}, layers(std::move(other.layers))
    {
    }
    ArvizIntrospector&
    ArvizIntrospector::operator=(ArvizIntrospector&&) noexcept
    {
        return *this;
    }
} // namespace armarx::nav::server
