#include "GraphScene.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <SemanticObjectRelations/Shapes/Shape.h>

#include <Eigen/Core>


namespace armarx::nav::locgrapheditor
{

    QGraphicsEllipseItem*
    GraphScene::addVertex(const graph::Graph::Vertex& vertex)
    {
        const Eigen::Matrix4f& pose = vertex.attrib().getPose();

        // To capture by copy
        semrel::ShapeID vertexID = vertex.objectID();

        GraphVisualizerGraphicsEllipseItem* item = new GraphVisualizerGraphicsEllipseItem
        {
            [this, vertexID]() { emit vertexSelected(vertexID); },
            static_cast<qreal>(pose(0, 3)),
            static_cast<qreal>(- pose(1, 3)),
            0, 0
        };
        addItem(item);

        // setToolTip on graphicsItem does not work
        item->setZValue(std::numeric_limits<qreal>::max());
        // dynamic_cast<QGraphicsItem*>(item)->setToolTip(QString::fromStdString("Vertex '" + name + "'"));
        item->setToolTip(QString::fromStdString("Vertex '" + vertex.attrib().getName() + "'"));

        return item;
    }


    QGraphicsLineItem*
    GraphScene::addEdge(const graph::Graph::Edge& edge)
    {
        semrel::ShapeID sourceID = edge.sourceObjectID();
        semrel::ShapeID targetID = edge.targetObjectID();

        Eigen::Matrix4d sourcePose = edge.source().attrib().getPose().cast<qreal>();
        Eigen::Matrix4d targetPose = edge.target().attrib().getPose().cast<qreal>();

        GraphVisualizerGraphicsLineItem* item = new GraphVisualizerGraphicsLineItem
        {
            [this, sourceID, targetID]() { emit edgeSelected(sourceID, targetID); },
            sourcePose(0, 3), - sourcePose(1, 3),
            targetPose(0, 3), - targetPose(1, 3)
        };
        addItem(item);

        // setToolTip on item does not work
        std::stringstream toolTip;
        toolTip << "Edge '" << edge.source().attrib().getName()
                << "' -> '"  << edge.target().attrib().getName();
        item->setToolTip(QString::fromStdString(toolTip.str()));

        return item;
    }


    void GraphScene::updateVertex(GuiGraph::Vertex& vertex)
    {
        QGraphicsEllipseItem* item = vertex.attrib().graphicsItem;
        ARMARX_CHECK_NOT_NULL(item);

        const Eigen::Matrix4d pose = vertex.attrib().getPose().cast<qreal>();

        QColor color = vertex.attrib().highlighted ? colorSelected : colorDefault;
        double lineWidth = vertex.attrib().highlighted ? lineWidthSelected : lineWidthDefault;

        item->setPen(QPen {color});
        item->setBrush(QBrush {color});
        item->setRect(  pose(0, 3) - lineWidth * verticesScaleFactor / 2,
                      - pose(1, 3) - lineWidth * verticesScaleFactor / 2,
                      lineWidth * verticesScaleFactor,
                      lineWidth * verticesScaleFactor);
    }


    void GraphScene::updateEdge(GuiGraph::Edge& edge)
    {
        QGraphicsLineItem* item = edge.attrib().graphicsItem;
        ARMARX_CHECK_NOT_NULL(item);

        Eigen::Matrix4d sourcePose = edge.source().attrib().getPose().cast<qreal>();
        Eigen::Matrix4d targetPose = edge.target().attrib().getPose().cast<qreal>();

        QColor color = edge.attrib().highlighted ? colorSelected : colorDefault;
        double lineWidth = edge.attrib().highlighted ? lineWidthSelected : lineWidthDefault;

        QPen pen {color};
        pen.setWidthF(lineWidth * lineScaleFactor);

        item->setPen(pen);
        item->setLine(sourcePose(0, 3), - sourcePose(1, 3),
                      targetPose(0, 3), - targetPose(1, 3));
    }


    void GraphScene::removeVertex(QGraphicsEllipseItem*& item)
    {
        removeItem(item);
        delete item;
        item = nullptr;
    }


    void GraphScene::removeEdge(QGraphicsLineItem*& item)
    {
        removeItem(item);
        delete item;
        item = nullptr;
    }


    GraphVisualizerGraphicsEllipseItem::GraphVisualizerGraphicsEllipseItem(
            std::function<void(void)> onDoubleClicked,
            qreal x, qreal y,
            qreal width, qreal height,
            QGraphicsItem* parent) :
        QGraphicsEllipseItem {x, y, width, height, parent},
        onDoubleClicked{onDoubleClicked}
    {
    }


    GraphVisualizerGraphicsEllipseItem::~GraphVisualizerGraphicsEllipseItem()
    {
    }


    void GraphVisualizerGraphicsEllipseItem::mouseDoubleClickEvent(QGraphicsSceneMouseEvent*)
    {
        onDoubleClicked();
    }



    GraphVisualizerGraphicsLineItem::GraphVisualizerGraphicsLineItem(
            std::function<void(void)> onDoubleClicked,
            qreal x1, qreal y1, qreal x2, qreal y2,
            QGraphicsItem* parent) :
        QGraphicsLineItem {x1, y1, x2, y2, parent},
        onDoubleClicked{onDoubleClicked}
    {
    }


    GraphVisualizerGraphicsLineItem::~GraphVisualizerGraphicsLineItem()
    {
    }


    void GraphVisualizerGraphicsLineItem::mouseDoubleClickEvent(QGraphicsSceneMouseEvent*)
    {
        onDoubleClicked();
    }

}

