/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http:// www.gnu.org/licenses/>.
 *
 * \package    Navigation::gui-plugins::LocationGraphEditorWidgetController
 * \author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * \date       2021
 * \copyright  http:// www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <VirtualRobot/VirtualRobot.h>

#include "LocationGraphEditorWidgetController.h"
#include "Visu.h"
#include "widgets/EdgeTableWidget.h"
#include "widgets/VertexDataWidget.h"
#include "widgets/VertexTableWidget.h"
#include "widgets/graph_scene/Scene.h"
#include "widgets/graph_scene/Widget.h"

#include <Navigation/gui-plugins/LocationGraphEditor/ui_LocationGraphEditorWidget.h>

#include <armarx/navigation/location/constants.h>
#include <armarx/navigation/graph/constants.h>
#include <armarx/navigation/graph/constants.h>
#include <armarx/navigation/graph/aron/Graph.aron.generated.h>

#include <RobotAPI/libraries/armem/client/MemoryNameSystem.h>
#include <RobotAPI/components/ArViz/Client/Client.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <SimoxUtility/color/Color.h>

// Qt headers
#include <QDialog>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QMenu>
#include <QMouseEvent>
#include <QObject>
#include <QPushButton>
#include <QSignalBlocker>

// std
#include <sstream>


static const QString SETTING_LAST_SCENE = "lastScene";


namespace armarx::nav::locgrapheditor
{


    /**
     * @brief Returns the name used on the debug layer.
     * @param edge The edge.
     * @return The name used on the debug layer.
     */
    std::string iceName(const GuiGraph::Edge& edge)
    {
        std::stringstream ss;
        ss << "edge " << std::to_string(edge.sourceObjectID()) << " -> " << std::to_string(edge.targetObjectID());
        return ss.str();
    }

    /**
     * @brief iceName Returns the name used on the debug layer.
     * @param vertexName The vertex.
     * @return The name used on the debug layer.
     */
    std::string iceName(const GuiGraph::Vertex& vertex)
    {
        return std::to_string(vertex.objectID());
    }


    QString LocationGraphEditorWidgetController::GetWidgetName()
    {
        return "Navigation.LocationGraphEditor";
    }
    QIcon LocationGraphEditorWidgetController::GetWidgetIcon()
    {
        return QIcon {":// icons/graph_visu.svg"};
    }


    LocationGraphEditorWidgetController::LocationGraphEditorWidgetController() :
        settings {"KIT", "LocationGraphEditorWidgetController"}
    {
        widget.setupUi(getWidget());

        loadAutomaticSettings();

        widget.loadGraphButton->setEnabled(false);

        view.edgeTable = new EdgeTableWidget();
        view.vertexTable = new VertexTableWidget();
        view.vertexData = new VertexDataWidget();
        view.vertexData->setEnabled(false);  // Enable on first selection of vertex.
        widget.edgesTableGroupBox->layout()->addWidget(view.edgeTable);
        widget.locationsTableGroupBox->layout()->addWidget(view.vertexTable);
        widget.locationDataGroupBox->layout()->addWidget(view.vertexData);

        view.graph = new GraphSceneWidget();
        widget.graphSceneLayout->addWidget(view.graph);


        // Widgets -> This

        // Memory Access
        connect(this, &This::connected, this, &This::queryGraphs);
        connect(this, &This::memoryDataChanged, this, &This::updateGraphList);
        connect(widget.refreshGraphsButton, &QPushButton::pressed, this, &This::updateGraphList);

        connect(widget.loadGraphButton, &QPushButton::pressed, this, &This::loadGraph);
        connect(widget.commitGraphButton, &QPushButton::pressed, this, &This::commitGraph);


        // Update views
        connect(this, &This::graphChanged, this, &This::updateGraphView);
        connect(view.vertexData, &VertexDataWidget::vertexDataChanged,
                this, &This::updateGraphView);

        // Selection
        connect(view.vertexTable, &VertexTableWidget::currentItemChanged,
                [this](QTableWidgetItem* current, QTableWidgetItem* previous)
        {
            (void) previous;
            this->selectVertex(current);
        });

        connect(view.vertexTable, &VertexTableWidget::itemSelectionChanged,
                this, &This::updateVertexHighlighting);
        connect(view.edgeTable, &EdgeTableWidget::itemSelectionChanged,
                this, &This::updateEdgeHighlighting);

        connect(view.vertexTable, &VertexTableWidget::newEdgesRequested,
                this, &This::addEdges);
        connect(view.edgeTable, &EdgeTableWidget::edgeRemovalRequested,
                this, &This::removeEdges);
    }


    LocationGraphEditorWidgetController::~LocationGraphEditorWidgetController()
    {
        saveAutomaticSettings();
    }


    QPointer<QDialog> LocationGraphEditorWidgetController::getConfigDialog(QWidget* parent)
    {
        if (not dialog)
        {
            dialog = new SimpleConfigDialog(parent);
            dialog->addProxyFinder<armem::mns::MemoryNameSystemInterfacePrx>("MemoryNameSystem", "Memory Name System", remote.memoryNameSystemName);
        }
        return qobject_cast<SimpleConfigDialog*>(dialog);
    }


    void LocationGraphEditorWidgetController::configured()
    {
        remote.memoryNameSystemName = dialog->getProxyName("MemoryNameSystem");
    }


    void LocationGraphEditorWidgetController::loadSettings(QSettings* settings)
    {
        remote.memoryNameSystemName = settings->value("memoryNameSystemName", QString::fromStdString(remote.memoryNameSystemName)).toString().toStdString();
    }


    void LocationGraphEditorWidgetController::saveSettings(QSettings* settings)
    {
        settings->setValue("memoryNameSystemName", QString::fromStdString(remote.memoryNameSystemName));
    }


    void LocationGraphEditorWidgetController::loadAutomaticSettings()
    {
        lastSelectedSceneName = settings.value(SETTING_LAST_SCENE, lastSelectedSceneName).toString();
    }


    void LocationGraphEditorWidgetController::saveAutomaticSettings()
    {
        settings.setValue(SETTING_LAST_SCENE, lastSelectedSceneName);
    }


    void LocationGraphEditorWidgetController::onInitComponent()
    {
        usingProxy(remote.memoryNameSystemName);
    }


    void LocationGraphEditorWidgetController::onConnectComponent()
    {
        remote.connect(*this);

        widget.sceneGroupBox->setEnabled(true);

        emit connected();
    }


    void LocationGraphEditorWidgetController::Remote::connect(Component& parent)
    {
        auto mnsProxy = parent.getProxy<armem::mns::MemoryNameSystemInterfacePrx>(memoryNameSystemName);
        memoryNameSystem = armem::client::MemoryNameSystem(mnsProxy, &parent);

        locationReader = memoryNameSystem.useReader(nav::loc::coreSegmentID);
        locationWriter = memoryNameSystem.useWriter(nav::loc::coreSegmentID);

        graphReader = memoryNameSystem.useReader(nav::graph::coreSegmentID);
        graphWriter = memoryNameSystem.useWriter(nav::graph::coreSegmentID);

        arviz = std::make_unique<viz::Client>(viz::Client::createForGuiPlugin(parent));
    }


    void LocationGraphEditorWidgetController::queryGraphs()
    {
        armem::client::QueryResult result = remote.graphReader.getLatestSnapshotsIn(nav::graph::coreSegmentID);
        if (result.success)
        {
            model.memory = std::move(result.memory);
            emit memoryDataChanged();
        }
        else
        {
            widget.statusLabel->setText(QString::fromStdString(result.errorMessage));
        }
    }


    void LocationGraphEditorWidgetController::updateGraphList()
    {
        QString previousText = widget.graphsComboBox->currentText();
        widget.graphsComboBox->clear();

        int i = 0;
        int previousIndex = -1;
        model.memory.forEachEntity([&](const armem::wm::Entity& entity)
        {
            QString text = QString::fromStdString(entity.id().str());
            widget.graphsComboBox->addItem(text);
            if (previousIndex < 0 and text == previousText)
            {
                previousIndex = i;
            }
            ++i;
        });
        if (previousIndex >= 0)
        {
            widget.graphsComboBox->setCurrentIndex(previousIndex);
        }
        widget.loadGraphButton->setEnabled(widget.graphsComboBox->count() > 0);
    }


    void LocationGraphEditorWidgetController::loadGraph()
    {
        const armem::MemoryID entityID = armem::MemoryID::fromString(widget.graphsComboBox->currentText().toStdString());

        queryGraphs();  // Refresh local memory.

        const armem::wm::EntityInstance* instance = model.memory.findLatestInstance(entityID);
        if (instance)
        {
            widget.statusLabel->setText(QString::fromStdString("Loaded snapshot " + instance->id().getEntitySnapshotID().str()));
        }
        else
        {
            std::stringstream ss;
            ss << "No latest instance of entity " << entityID << " in memory.";
            widget.statusLabel->setText(QString::fromStdString(ss.str()));
        }

        nav::graph::Graph nav;
        {
            nav::graph::arondto::Graph dto;
            dto.fromAron(instance->data());
            fromAron(dto, nav);
        }

        model.graphEntityID = entityID;
        ARMARX_VERBOSE << "Loading graph " << nav.str();
        setGraph(nav);
    }


    void LocationGraphEditorWidgetController::commitGraph()
    {
        nav::graph::arondto::Graph dto;
        {
            nav::graph::Graph nav = fromGuiGraph(model.graph);
            toAron(dto, nav);
        }

        armem::EntityUpdate update;
        update.entityID = model.graphEntityID;
        update.timeCreated = armem::Time::now();
        update.instancesData = {dto.toAron()};

        armem::EntityUpdateResult result = remote.graphWriter.commit(update);
        if (result.success)
        {
            widget.statusLabel->setText(QString::fromStdString("Committed snapshot " + result.snapshotID.str()));
        }
        else
        {
            widget.statusLabel->setText(QString::fromStdString(result.errorMessage));
        }
    }


    void LocationGraphEditorWidgetController::setGraph(graph::Graph& nav)
    {
#if 0
        // Store vertex highlighting.
        std::set<std::string> highlightedVertices;
        for (auto vertex : model.graph.vertices())
        {
            if (vertex.attrib().highlighted)
            {
                highlightedVertices.insert(vertex.attrib().getName());
            }
        }
#endif

        // Build the gui graph (model).
        {
            QSignalBlocker blocker(this);

            clearGraph();
            for (auto vertex : nav.vertices())
            {
                addVertex(vertex);
            }
            for (auto edge : nav.edges())
            {
                addEdge(edge);
            }

#if 0
            // Restore vertex highlighting.
            for (auto vertex : model.graph.vertices())
            {
                if (highlightedVertices.count(vertex.attrib().getName()))
                {
                    vertex.attrib().highlighted = true;
                }
            }
#endif

        }

        // Trigger a view update.
        emit graphChanged();
    }


    GuiGraph::Vertex
    LocationGraphEditorWidgetController::addVertex(graph::Graph::ConstVertex vertex)
    {
        ARMARX_CHECK(not model.graph.hasVertex(vertex.objectID()));

        VertexData attrib { vertex.attrib() };
        attrib.graphicsItem = view.graph->scene()->addVertex(vertex);
        attrib.tableWidgetItem = view.vertexTable->addVertex(vertex);
        return model.graph.addVertex(vertex.objectID(), attrib);
    }


    GuiGraph::Edge
    LocationGraphEditorWidgetController::addEdge(graph::Graph::ConstEdge edge)
    {
        auto source = model.graph.vertex(edge.sourceObjectID());
        auto target = model.graph.vertex(edge.targetObjectID());

        return addEdge(source, target, { edge.attrib() });
    }


    GuiGraph::Edge
    LocationGraphEditorWidgetController::addEdge(
            GuiGraph::ConstVertex source,
            GuiGraph::ConstVertex target,
            const EdgeData& defaultAttribs)
    {
        ARMARX_CHECK(not model.graph.hasEdge(source.objectID(), target.objectID()))
                << "Edge must not exist before being added: '"
                << source.attrib().getName() << "' -> '"
                << target.attrib().getName() << "'";

        GuiGraph::Edge edge = model.graph.addEdge(source, target, defaultAttribs);
        edge.attrib().tableWidgetItem = view.edgeTable->addEdge(edge);
        edge.attrib().graphicsItem = view.graph->scene()->addEdge(edge);
        return edge;
    }


    void LocationGraphEditorWidgetController::addEdges(
            QList<QPair<QTableWidgetItem*, QTableWidgetItem*>> vertexItems)
    {
        if (/* DISABLES CODE */ (false))
        {
            std::stringstream ss;
            ss << "Adding edges ...";
            for (const auto& [source, target] : vertexItems)
            {
                ss << "\n- " << source->text() << " to " << target->text();
            }
            ARMARX_IMPORTANT << ss.str();
        }

        const auto itemToVertexMap = model.graph.getTableItemToVertexMap();
        for (const auto& [sourceItem, targetItem] : vertexItems)
        {
            GuiGraph::Vertex source = itemToVertexMap.at(sourceItem);
            GuiGraph::Vertex target = itemToVertexMap.at(targetItem);
            if (not model.graph.hasEdge(source, target))
            {
                addEdge(source, target, {});
            }
        }

        emit graphChanged();
    }


    void LocationGraphEditorWidgetController::removeEdges(
            QList<QTableWidgetItem*> edgeItems)
    {
        if (/* DISABLES CODE */ (true))
        {
            std::stringstream ss;
            ss << "Remoiving edges ...";
            for (const auto& edgeItem : edgeItems)
            {
                ss << "\n- " << edgeItem->text();
            }
            ARMARX_IMPORTANT << ss.str();
        }

        {
            QSignalBlocker blocker(this);

            const auto itemToEdgeMap = model.graph.getTableItemToEdgeMap();
            for (const auto& edgeItem : edgeItems)
            {
                GuiGraph::Edge edge = itemToEdgeMap.at(edgeItem);
                removeEdge(edge);
            }
        }

        emit graphChanged();
    }


    void LocationGraphEditorWidgetController::updateGraphView()
    {
        for (auto vertex : model.graph.vertices())
        {
            updateVertexView(vertex);
        }
        for (auto edge : model.graph.edges())
        {
            updateEdgeView(edge);
        }
        updateArViz();
    }


    void
    LocationGraphEditorWidgetController::updateVertexView(GuiGraph::Vertex vertex)
    {
        view.graph->scene()->updateVertex(vertex);
        view.vertexTable->updateVertex(vertex);

        if (/* DISABLES CODE */ (false))  // Disable this for the moment.
        {
            // Highlight all edges between highlighted vertices
            std::set<semrel::ShapeID> highlightedVertices;
            for (auto v : model.graph.vertices())
            {
                if (v.attrib().highlighted)
                {
                    highlightedVertices.insert(v.objectID());
                }
            }

            for (auto edge : model.graph.edges())
            {
                bool verticesHighlighted =
                        highlightedVertices.count(edge.sourceObjectID())
                        and highlightedVertices.count(edge.targetObjectID());

                // Already highlighted but vertices not highlighted => to false, update
                // Not highlighted but vertices highlighted => to true, update
                if (edge.attrib().highlighted != verticesHighlighted)
                {
                    edge.attrib().highlighted = verticesHighlighted;
                }
            }
        }
    }


    void
    LocationGraphEditorWidgetController::updateEdgeView(GuiGraph::Edge edge)
    {
        view.graph->scene()->updateEdge(edge);
        view.edgeTable->updateEdge(edge);
    }


    void LocationGraphEditorWidgetController::updateArViz()
    {
        if (remote.arviz)
        {
            viz::Layer layer = remote.arviz->layer(widget.graphsComboBox->currentText().toStdString());
            GraphVisu visu;
            visu.draw(layer, model.graph);
            remote.arviz->commit(layer);
        }
    }


    void LocationGraphEditorWidgetController::clearGraph()
    {
        {
            QSignalBlocker blocker(this);
            clearEdges();
            clearVertices();
        }

        // Clear data structure
        ARMARX_CHECK_EQUAL(model.graph.numEdges(), 0);
        ARMARX_CHECK_EQUAL(model.graph.numVertices(), 0);

        emit graphChanged();
    }


    void LocationGraphEditorWidgetController::clearEdges()
    {
        // Remove in reverse order to be more array-friendly.
        std::vector<GuiGraph::Edge> edges { model.graph.edges().begin(), model.graph.edges().end() };
        for (auto it = edges.rbegin(); it != edges.rend(); ++it)
        {
            GuiGraph::Edge edge = *it;
            removeEdge(edge);
        }

        ARMARX_CHECK_EQUAL(view.edgeTable->rowCount(), 0);
        ARMARX_CHECK_EQUAL(model.graph.numEdges(), 0);

        emit graphChanged();
    }


    void LocationGraphEditorWidgetController::clearVertices()
    {
        ARMARX_CHECK_EQUAL(model.graph.numEdges(), 0)
                << "The graph may not have any edges when clearing the vertices.";

        // Remove in reverse order to be more array-friendly.
        std::vector<GuiGraph::Vertex> vertices { model.graph.vertices().begin(), model.graph.vertices().end() };
        for (auto it = vertices.rbegin(); it != vertices.rend(); ++it)
        {
            GuiGraph::Vertex vertex = *it;
            removeVertex(vertex);
        }

        ARMARX_CHECK_EQUAL(view.vertexTable->rowCount(), 0);
        ARMARX_CHECK_EQUAL(model.graph.numVertices(), 0);

        emit graphChanged();
    }


    void LocationGraphEditorWidgetController::removeEdge(GuiGraph::Edge& edge)
    {
        ARMARX_CHECK(model.graph.hasEdge(edge.sourceDescriptor(), edge.targetDescriptor()))
                << "Cannot remove edge that does not exist.";

        // Remove view elements
        view.graph->scene()->removeEdge(edge.attrib().graphicsItem);

        const int rows = view.edgeTable->rowCount();
        view.edgeTable->removeEdge(edge);
        ARMARX_CHECK_EQUAL(view.edgeTable->rowCount(), rows - 1);

        // Remove from model
        model.graph.removeEdge(edge);

        ARMARX_CHECK(not model.graph.hasEdge(edge.sourceDescriptor(), edge.targetDescriptor()))
                << edge.sourceDescriptor() << " -> " << edge.targetDescriptor();
    }


    void LocationGraphEditorWidgetController::removeVertex(GuiGraph::Vertex& vertex)
    {
        ARMARX_CHECK_EQUAL(vertex.inDegree(), 0)
                << "A vertex may not have any edges before being removed. " << vertex.attrib().getName();
        ARMARX_CHECK_EQUAL(vertex.outDegree(), 0)
                << "A vertex may not have any edges before being removed. "  << vertex.attrib().getName();

        // Remove view elements
        view.graph->scene()->removeVertex(vertex.attrib().graphicsItem);
        view.vertexTable->removeVertex(vertex);
        if (view.vertexData->vertex().has_value() and view.vertexData->vertex().value() == vertex)
        {
            view.vertexData->clearVertex();
        }

        // Remove from model
        model.graph.removeVertex(vertex);
    }


    void LocationGraphEditorWidgetController::resetHighlighting()
    {
        for (auto vertex : model.graph.vertices())
        {
            if (vertex.attrib().highlighted)
            {
                vertex.attrib().highlighted = false;
            }
        }
        for (auto edge : model.graph.edges())
        {
            if (edge.attrib().highlighted)
            {
                edge.attrib().highlighted = false;
            }
        }

        emit graphChanged();
    }


    template <class T>
    static
    void updateElementHighlighting(
            QList<QTableWidgetItem*> selectedItems,
            std::map<QTableWidgetItem*, T>&& itemToElementMap)
    {
        for (QTableWidgetItem* selected : selectedItems)
        {
            if (auto it = itemToElementMap.find(selected); it != itemToElementMap.end())
            {
                it->second.attrib().highlighted = true;
                itemToElementMap.erase(it);
            }
        }
        for (auto& [_, unselected] : itemToElementMap)
        {
            unselected.attrib().highlighted = false;
        }
    }


    void LocationGraphEditorWidgetController::updateVertexHighlighting()
    {
        updateElementHighlighting(view.vertexTable->selectedVertexItems(), model.graph.getTableItemToVertexMap());
        emit graphChanged();
    }


    void LocationGraphEditorWidgetController::updateEdgeHighlighting()
    {
        updateElementHighlighting(view.edgeTable->selectedEdgeItems(), model.graph.getTableItemToEdgeMap());
        emit graphChanged();
    }


    void LocationGraphEditorWidgetController::selectVertex(QTableWidgetItem* vertexItem)
    {
        if (vertexItem == nullptr)
        {
            view.vertexData->clearVertex();
        }
        if (auto vertex = model.graph.getVertexFromTableItem(vertexItem))
        {
            selectVertex(vertex.value());
        }
    }


    void LocationGraphEditorWidgetController::selectVertex(GuiGraph::Vertex vertex)
    {
        view.vertexData->setVertex(vertex);
    }

}
