/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::gui-plugins::LocationGraphEditorWidgetController
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#pragma once

#include "FunctionalEventFilter.h"
#include "GuiGraph.h"
#include "widgets/graph_scene.h"

#include <armarx/navigation/graph/Graph.h>

#include <Navigation/gui-plugins/LocationGraphEditor/ui_LocationGraphEditorWidget.h>

#include <RobotAPI/libraries/armem/core/wm/memory_definitions.h>
#include <RobotAPI/libraries/armem/client/forward_declarations.h>
#include <RobotAPI/libraries/armem/client/MemoryNameSystem.h>
#include <RobotAPI/libraries/armem/client/Reader.h>
#include <RobotAPI/libraries/armem/client/Writer.h>

#include <ArmarXGui/libraries/ArmarXGuiBase/ArmarXComponentWidgetController.h>
#include <ArmarXGui/libraries/SimpleConfigDialog/SimpleConfigDialog.h>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/system/ImportExportComponent.h>

#include <memory>
#include <string>


class QDialog;


namespace armarx::viz
{
    class Client;
}
namespace armarx::nav::locgrapheditor
{
    class EdgeTableWidget;
    class VertexDataWidget;
    class VertexTableWidget;


    /**
    \page Navigation-GuiPlugins-LocationGraphEditor LocationGraphEditor
    \brief The LocationGraphEditor allows visualizing ...

    \image html LocationGraphEditor.png
    The user can

    API Documentation \ref LocationGraphEditorWidgetController

    \see LocationGraphEditorGuiPlugin
    */


    /**
     * \class LocationGraphEditorGuiPlugin
     * \ingroup ArmarXGuiPlugins
     * \brief LocationGraphEditorGuiPlugin brief description
     *
     * Detailed description
     */

    /**
     * \class LocationGraphEditorWidgetController
     * \brief LocationGraphEditorWidgetController brief one line description
     *
     * Detailed description
     */
    class ARMARXCOMPONENT_IMPORT_EXPORT
        LocationGraphEditorWidgetController :
            public armarx::ArmarXComponentWidgetControllerTemplate < LocationGraphEditorWidgetController >
    {
        Q_OBJECT

        using This = LocationGraphEditorWidgetController;


    public:

        static QString GetWidgetName();
        static QIcon GetWidgetIcon();


        explicit LocationGraphEditorWidgetController();
        virtual ~LocationGraphEditorWidgetController() override;


        //// @see ArmarXWidgetController::loadSettings()
        void loadSettings(QSettings* settings) override;
        /// @see ArmarXWidgetController::saveSettings()
        void saveSettings(QSettings* settings) override;


        QPointer<QDialog> getConfigDialog(QWidget* parent = nullptr) override;
        void configured() override;


        void onInitComponent() override;
        void onConnectComponent() override;


        //because the above load/save functions are *obviously* for "Save/Load Gui Config." :/
        virtual void loadAutomaticSettings();
        virtual void saveAutomaticSettings();


    signals:

        void connected();
        void memoryDataChanged();
        void graphChanged();


    private slots:

        // Model

        void queryGraphs();
        void updateGraphList();
        void loadGraph();
        void commitGraph();

        void setGraph(graph::Graph& nav);
        GuiGraph::Vertex addVertex(graph::Graph::ConstVertex vertex);
        GuiGraph::Edge addEdge(graph::Graph::ConstEdge edge);
        GuiGraph::Edge addEdge(GuiGraph::ConstVertex source, GuiGraph::ConstVertex target,
                               const EdgeData& defaultAttribs);

        void clearGraph();
        void clearEdges();
        void clearVertices();

        void removeEdge(GuiGraph::Edge& edge);
        void removeVertex(GuiGraph::Vertex& vertex);

        void addEdges(QList<QPair<QTableWidgetItem*, QTableWidgetItem*>> vertexItems);
        void removeEdges(QList<QTableWidgetItem*> edgeItems);


        // View & Tables

        void updateGraphView();
        void updateVertexView(GuiGraph::Vertex vertex);
        void updateEdgeView(GuiGraph::Edge edge);
        void updateArViz();


        // Selection

        void selectVertex(QTableWidgetItem* vertexItem);
        void selectVertex(GuiGraph::Vertex vertex);

        void resetHighlighting();
        void updateVertexHighlighting();
        void updateEdgeHighlighting();


    private:

        /// Widget Form
        Ui::LocationGraphEditorWidget widget;
        QPointer<SimpleConfigDialog> dialog;


        struct Remote
        {
            std::string memoryNameSystemName = "MemoryNameSystem";
            armem::client::MemoryNameSystem memoryNameSystem;

            armem::client::Reader locationReader;
            armem::client::Writer locationWriter;

            armem::client::Reader graphReader;
            armem::client::Writer graphWriter;

            std::unique_ptr<viz::Client> arviz;

            void connect(Component& parent);
        };
        Remote remote;


        struct Model
        {
            armem::wm::Memory memory;
            armem::MemoryID graphEntityID;
            GuiGraph graph;
        };
        Model model;


        struct View
        {
            VertexTableWidget* vertexTable = nullptr;
            EdgeTableWidget* edgeTable = nullptr;
            VertexDataWidget* vertexData = nullptr;

            GraphSceneWidget* graph = nullptr;
        };
        View view;


    // Non-refactored

#if 0
    private slots:

        void tableWidgetVerticesCustomContextMenu(QPoint pos);
        void tableWidgetEdgesCustomContextMenu(QPoint pos);


    private slots:

        /// Add kitchen graph (H2T Armar3a robot kitchen)
        void addKitchenGraph();

        bool addNewEdge(const std::string& from, const std::string& to);
        void addNewEdgeBoth();
        void addNewEdgeStartEnd();
        void addNewEdgeEndStart();

        void addNewGraphVertex();
        void editGraphVertex();
#endif


    private:

        QSettings settings;
        QString lastSelectedSceneName;
    };

}

namespace armarx
{
    using armarx::nav::locgrapheditor::LocationGraphEditorWidgetController;
}

