/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    MemoryX::ArmarXObjects::GraphImportExport
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "VertexDataWidget.h"

#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/libraries/armem/core/aron_conversions.h>
#include <RobotAPI/libraries/aron/common/aron_conversions/core.h>

#include <QDoubleSpinBox>
#include <QFormLayout>
#include <QHBoxLayout>
#include <QLineEdit>
#include <QList>
#include <QRadioButton>
#include <QSignalBlocker>

#include <SimoxUtility/math/convert/deg_to_rad.h>
#include <SimoxUtility/math/convert/mat4f_to_rpy.h>
#include <SimoxUtility/math/convert/rad_to_deg.h>
#include <SimoxUtility/math/convert/rpy_to_mat4f.h>
#include <SimoxUtility/math/pose/pose.h>


namespace armarx::nav::locgrapheditor
{

    VertexDataWidget::VertexDataWidget()
    {
        setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Minimum);

        locationID = new QLineEdit(this);
        locationID->setReadOnly(true);

        frame = new QLineEdit(this);

        x = new QDoubleSpinBox(this);
        y = new QDoubleSpinBox(this);
        z = new QDoubleSpinBox(this);

        roll = new QDoubleSpinBox(this);
        pitch = new QDoubleSpinBox(this);
        yaw = new QDoubleSpinBox(this);

        angleUnitDeg = new QRadioButton("Degrees");
        angleUnitRad = new QRadioButton("Radians");


        for (QDoubleSpinBox* pos : _positionSpinBoxes())
        {
            pos->setSuffix(" mm");
            pos->setMinimum(-1e6);
            pos->setMaximum(+1e6);
            pos->setSingleStep(50);
        }
        for (QDoubleSpinBox* spinBox : _allSpinBoxes())
        {
            spinBox->setValue(0);
            spinBox->setAlignment(Qt::AlignRight);
        }

        QFormLayout* layout = new QFormLayout();
        this->setLayout(layout);

        layout->addRow("Location ID", locationID);
        layout->addRow("Frame", frame);
        layout->addRow("X", x);
        layout->addRow("Y", y);
        layout->addRow("Z", z);
        layout->addRow("Roll", roll);
        layout->addRow("Pitch", pitch);
        layout->addRow("Yaw", yaw);

        QHBoxLayout* angleUnitLayout = new QHBoxLayout();
        for (QRadioButton* angleUnit : {angleUnitDeg, angleUnitRad})
        {
            angleUnitLayout->addWidget(angleUnit);
        }
        layout->addRow("", angleUnitLayout);


        // Connect

        for (QRadioButton* angleUnit : {angleUnitDeg, angleUnitRad})
        {
            connect(angleUnit, &QRadioButton::toggled,
                    this, &This::_updateAngleUnit);
        }

        connect(frame, &QLineEdit::editingFinished,
                this, &This::_updateVertexAttribs);
        for (QDoubleSpinBox* spinBox : _allSpinBoxes())
        {
            connect(spinBox, QOverload<qreal>::of(&QDoubleSpinBox::valueChanged),
                    this, &This::_updateVertexAttribs);
        }


        angleUnitDeg->click();
    }


    std::optional<GuiGraph::Vertex> VertexDataWidget::vertex()
    {
        return _vertex;
    }


    void VertexDataWidget::setVertex(GuiGraph::Vertex vertex)
    {
        QSignalBlocker blocker(this);

        _vertex = vertex;
        _setFromVertex(vertex);
        setEnabled(true);
    }


    void VertexDataWidget::clearVertex()
    {
        _vertex = std::nullopt;
        setEnabled(false);
    }


    void VertexDataWidget::_setFromVertex(const GuiGraph::Vertex& vertex)
    {
        const VertexData& attrib = vertex.attrib();
        const Eigen::Matrix4d pose = attrib.getPose().cast<qreal>();

        locationID->setText(QString::fromStdString(aron::fromAron<armem::MemoryID>(attrib.aron.locationID).str()));
        frame->setText("<WIP>");
        _setXyz(simox::math::position(pose));
        _setRpyRad(simox::math::mat4f_to_rpy(pose));
    }


    void VertexDataWidget::_getToVertex(GuiGraph::Vertex& vertex)
    {
        VertexData& attrib = vertex.attrib();

        // locationID is read-only
        // frame->setText("<WIP>");  // WIP

        Eigen::Matrix4d pose = simox::math::pose(xyz(), simox::math::rpy_to_mat3f(rpyRad()));
        attrib.setPose(pose.cast<float>());
    }


    Eigen::Vector3d VertexDataWidget::xyz() const
    {
        return { x->value(), y->value(), z->value() };
    }


    Eigen::Vector3d VertexDataWidget::rpyDeg() const
    {
        Eigen::Vector3d raw = _rpyRaw();
        if (angleUnitRad->isChecked())
        {
            return {}; // return simox::math::rad_to_deg(raw);
        }
        else
        {
            return raw;
        }
    }


    Eigen::Vector3d VertexDataWidget::rpyRad() const
    {
        Eigen::Vector3d raw = _rpyRaw();
        if (angleUnitDeg->isChecked())
        {
            return simox::math::deg_to_rad(raw);
        }
        else
        {
            return raw;
        }
    }


    void VertexDataWidget::_updateAngleUnit()
    {
        std::function<double(double)> convertValue;
        QString suffix = " \u00b0";
        double min = -360;
        double max = +360;
        double step = 5.;
        int decimals = 2;

        if (angleUnitRad->isChecked())
        {
            convertValue = [](double deg)
            {
                return simox::math::deg_to_rad(deg);
            };
            suffix = " rad";
            min = simox::math::deg_to_rad(min);
            max = simox::math::deg_to_rad(max);
            step = simox::math::deg_to_rad(step);
            decimals = 3;
        }
        else
        {
            ARMARX_CHECK(angleUnitDeg->isChecked());
            convertValue = [](double rad)
            {
                return simox::math::rad_to_deg(rad);
            };
        }

        QList<QSignalBlocker>> blockers;
        for (QDoubleSpinBox* angle : _angleSpinBoxes())
        {
            blockers.append(QSignalBlocker(angle));

            angle->setSuffix(suffix);
            angle->setMinimum(min);
            angle->setMaximum(max);
            angle->setSingleStep(step);
            angle->setDecimals(decimals);
        }
        if (_vertex.has_value())
        {
            _setRpyRad(simox::math::mat4f_to_rpy(_vertex->attrib().getPose().cast<qreal>()));
        }

        // blockers will disable blocking for all spin boxes on destruction.
    }


    void VertexDataWidget::_updateVertexAttribs()
    {
        if (_vertex.has_value())
        {
            _getToVertex(_vertex.value());
            emit vertexDataChanged();
        }
    }


    std::vector<QDoubleSpinBox*> VertexDataWidget::_positionSpinBoxes()
    {
        return { x, y, z };
    }


    std::vector<QDoubleSpinBox*> VertexDataWidget::_angleSpinBoxes()
    {
        return { roll, pitch, yaw };
    }


    std::vector<QDoubleSpinBox*> VertexDataWidget::_allSpinBoxes()
    {
        return { x, y, z, roll, pitch, yaw };
    }


    void VertexDataWidget::_setXyz(const Eigen::Vector3d& xyz)
    {
        x->setValue(xyz.x());
        y->setValue(xyz.y());
        z->setValue(xyz.z());
    }


    Eigen::Vector3d VertexDataWidget::_rpyRaw() const
    {
        return { roll->value(), pitch->value(), yaw->value() };
    }


    void VertexDataWidget::_setRpyRaw(const Eigen::Vector3d& rpy) const
    {
        roll->setValue(rpy(0));
        pitch->setValue(rpy(1));
        yaw->setValue(rpy(2));
    }


    void VertexDataWidget::_setRpyDeg(const Eigen::Vector3d& rpyDeg) const
    {
        if (angleUnitDeg->isChecked())
        {
            _setRpyRaw(rpyDeg);
        }
        else if (angleUnitRad->isChecked())
        {
            _setRpyRaw(simox::math::deg_to_rad(rpyDeg));
        }
    }


    void VertexDataWidget::_setRpyRad(const Eigen::Vector3d& rpyRad) const
    {
        if (angleUnitDeg->isChecked())
        {
            _setRpyRaw(simox::math::rad_to_deg(rpyRad));
        }
        else if (angleUnitRad->isChecked())
        {
            _setRpyRaw(rpyRad);
        }
    }

}
