/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "VertexTableWidget.h"
#include "utils.h"

#include <QAction>
#include <QHeaderView>
#include <QMenu>


namespace armarx::nav::locgrapheditor
{

    VertexTableWidget::VertexTableWidget()
    {
        QStringList columns{"Name", "X [mm]", "Y [mm]", "Yaw [\u00b0]"};
        setColumnCount(columns.size());
        setHorizontalHeaderLabels(columns);
        horizontalHeader()->setResizeMode(0, QHeaderView::Stretch);
        horizontalHeader()->setVisible(true);

        setEditTriggers(QAbstractItemView::NoEditTriggers);
        setSortingEnabled(true);

        setAlternatingRowColors(true);

        QString styleSheet = this->styleSheet();
        styleSheet = styleSheet + "\n" + "selection-background-color: #FF8000;";
        setStyleSheet(styleSheet);

        setContextMenuPolicy(Qt::CustomContextMenu);
        connect(this, &This::customContextMenuRequested,
                this, &This::makeContextMenu);
    }


    QTableWidgetItem*
    VertexTableWidget::addVertex(graph::Graph::ConstVertex vertex)
    {
        (void) vertex;

        int row = rowCount();
        setRowCount(row + 1);

        // Just fill with vanilla items, they will get values in the update.
        for (int col = 0; col < 4; ++col)
        {
            setItem(row, col, new QTableWidgetItem {});
        }
        for (int col = 1; col <= 3; ++col)
        {
            item(row, col)->setTextAlignment(Qt::AlignRight | Qt::AlignVCenter);
        }

        return item(row, 0);
    }


    void
    VertexTableWidget::updateVertex(GuiGraph::Vertex vertex)
    {
        const Eigen::Matrix4d pose = vertex.attrib().getPose().cast<qreal>();
        char format = 'f';
        const int precision = 2;

        int row = this->row(vertex.attrib().tableWidgetItem);
        item(row, 0)->setText(QString::fromStdString(vertex.attrib().getName()));
        item(row, 1)->setText(QString::number(pose(0, 3), format, precision));
        item(row, 2)->setText(QString::number(pose(1, 3), format, precision));
        item(row, 3)->setText(QString::number(getYawAngleDegree(pose), format, precision));


        QColor bgColor = vertex.attrib().highlighted ? bgColorSelected : bgColorDefault;
        QFont font;
        font.setBold(vertex.attrib().highlighted);
        for (int col = 0; col < 4; ++col)
        {
            auto* item = this->item(row, col);
            ARMARX_CHECK_NOT_NULL(item);

            item->setData(Qt::BackgroundRole, bgColor);
            item->setFont(font);
        }
    }


    void
    VertexTableWidget::removeVertex(GuiGraph::Vertex& vertex)
    {
        if (currentItem() == vertex.attrib().tableWidgetItem)
        {
            setCurrentItem(nullptr);
        }
        this->removeRow(row(vertex.attrib().tableWidgetItem));
        vertex.attrib().tableWidgetItem = nullptr;
    }


    QList<QTableWidgetItem*>
    VertexTableWidget::selectedVertexItems()
    {
        return utils::getSelectedItemsOfColumn(this, 0);
    }


    void VertexTableWidget::makeContextMenu(QPoint pos)
    {
        QList<QTableWidgetItem*> items = selectedVertexItems();

        QMenu menu;
        if (items.size() == 0)
        {
            QAction* action = menu.addAction("No locations selected");
            action->setEnabled(false);
        }

        // Partners selected
        if (items.size() == 2)
        {
            menu.addSection("Selected pair of locations");

            // Generate actions for connecting these two nodes.
            std::sort(items.begin(), items.end(), [](QTableWidgetItem* first, QTableWidgetItem* second)
            {
                return first->text() < second->text();
            });
            QTableWidgetItem* first = items[0];
            QTableWidgetItem* second = items[1];

            connect(menu.addAction("Add edge '" + first->text() + "' " + utils::arrowRight + " '" + second->text() + "'"),
                    &QAction::triggered, [this, first, second]()
            {
                emit newEdgesRequested({{first, second}});
            });
            connect(menu.addAction("Add edge '" + first->text() + "' " + utils::arrowLeft + " '" + second->text() + "'"),
                    &QAction::triggered, [this, first, second]()
            {
                emit newEdgesRequested({{second, first}});
            });
            connect(menu.addAction("Add edges '" + first->text() + "' " + utils::arrowBoth + " '" + second->text() + "'"),
                    &QAction::triggered, [this, first, second]()
            {
                emit newEdgesRequested({{first, second}, {second, first}});
            });
        }

        // Partners via context menu
        if (items.size() > 0)
        {
            QString edges = items.size() == 1
                    ? "edge"
                    : "edges";
            QString desc = items.size() == 1
                    ? "'" + items[0]->text() + "'"
                    : QString::number(items.size()) + " locations";

            if (items.size() == 1)
            {
                // QAction* deleteAction = menu.addAction("Delete location '" + );
                menu.addSection("Selected single location " + desc);
            }
            else
            {
                menu.addSection("Selected bulk of " + desc);
            }

            using Item = QTableWidgetItem;
            using ListOfEdges = QList<QPair<Item*, Item*>>;
            using AppendFunc = std::function<void(ListOfEdges& edges, Item* selected, Item* action)>;

            auto addBulkActions = [this, &items](QMenu* submenu, AppendFunc appendFunc)
            {
                if (items.size() == rowCount())
                {
                    QAction* a = submenu->addAction("No other locations");
                    a->setDisabled(true);
                }
                for (int row = 0; row < rowCount(); ++row)
                {
                    QTableWidgetItem* action = this->item(row, 0);
                    if (items.count(action) == 0)  // Do no generate self-edges
                    {
                        QAction* a = submenu->addAction(action->text());
                        connect(a, &QAction::triggered,
                                [this, &items, action, appendFunc]()
                        {
                            QList<QPair<QTableWidgetItem*, QTableWidgetItem*>> edges;
                            for (auto* selected : items)
                            {
                                appendFunc(edges, selected, action);
                            }
                            emit newEdgesRequested(edges);
                        });
                    }
                }
            };

            addBulkActions(menu.addMenu("Add " + edges + " " + desc + " " + utils::arrowRight + " ..."),
                           [](ListOfEdges& edges, Item* selected, Item* action)
            {
                edges.append({selected, action});
            });
            addBulkActions(menu.addMenu("Add " + edges + " " + desc + " " + utils::arrowLeft + " ..."),
                           [](ListOfEdges& edges, Item* selected, Item* action)
            {
                edges.append({action, selected});
            });
            addBulkActions(menu.addMenu("Add " + edges + " " + desc + " " + utils::arrowBoth + " ..."),
                           [](ListOfEdges& edges, Item* selected, Item* action)
            {
                edges.append({selected, action});
                edges.append({action, selected});
            });
        }

        menu.exec(mapToGlobal(pos));
    }

}
